/***********************************************************************

------------
INTRODUCTION
------------

This  is  a  collection  of  FORTRAN-callable C-style I/O routines, a la
4.2BSD and AT&T System V Unix.	 The  current  implementation  is  under
4.1BSD  Eunice  running under VAX VMS.  Section 3 of the Unix manual may
be consulted for a precise definition of what  each  of  these  routines
does.	When  reading the manual sections, replace references to C FILE*
pointers and integer file descriptors by FORTRAN unit numbers.

The 4.2 BSD library in section 3F contains:

	getc, fgetc, putc, fputc, fflush, perror

The argument order in fgetc() and fputc() is reversed from that  in  the
standard  C  functions; somebody was not thinking clearly when these two
were defined, sigh...  In all of the  other  functions  below,  argument
order is the SAME in both FORTRAN and C versions.

We have taken the liberty here of providing the additional ones:

	clrerr
	getw, putw
	fread, fwrite
	fcread, fcwrit
	feof, ferror
	fseek, ftell
	isatty
	strlen
	ungetc

which  have  similar  behavior  to  the standard C functions.  Functions
which return char* pointers (fgets() and gets()) have  been  changed  to
return 0 on success, and EOF on error.

Notably absent from this list are fopen(), fclose(), and rewind(), which
are  assumed  to  be  supplied  by  FORTRAN  OPEN,  CLOSE,  and   REWIND
statements.

Also  absent  are  fprintf(),  fscanf(),  printf(),  and  scanf(), which
require variable-length argument lists, and are  notorious  barriers  to
portability  of  the  C  library.   All  C  implementations  can support
functions with variable numbers of arguments, so it is regrettable  that
a  standard  mechanism for doing this was not defined into the language.
As it is now, implementations require knowledge  of  argument  order  to
construct  a  vector  of  argument pointers, which are then passed on to
non-standard lower-level routines which perform the necessary functions.

clrerr() maps  to  clearerr()  because  of  FORTRAN's  6-character  name
limitation.   fcread()  and  fcwrit()  are  equivalent  to  fread()  and
fwrite(), except that they take  a  CHARACTER  argument.   Unix  FORTRAN
compilers  based  on  the original f77 (including 4.1BSD and 4.2BSD f77)
generally permit equivalencing of CHARACTER and  non-CHARACTER  data  by
EQUIVALENCE  statements  or  argument  associations,  but  several newer
compilers (e.g.  fc on AT&T System V, HPUX, and CONVEX), do not,  so  we
must offer two variants of these routines.

A  choice  must  be  made as to what constitutes a "string" for fgets(),
fputs(), gets(), and  puts().	FORTRAN  has  blank-padded  fixed-length
strings, while C has NUL-terminated variable-length strings.  Because it
is anticipated that these will be used precisely to  get  at  the  exact
contents  of  files, blank padding is undesirable, and a normal C string
will be returned, provided there is room for the  NUL  terminator.   The
strlen()  function  can  be  used to obtain the length of such a C-style
string.  If there is no NUL  terminator  within  the  first  LEN(STRING)
characters, then STRING(LEN(STRING)+1) will be implicitly treated as NUL
for these routines.

Efficiency has  been  a  prime  consideration  in  the  design  of  this
interface,  and  heavy  use  has been made of register declarations, and
preference given to in-line macro forms instead of function calls.


-----
USAGE
-----
	SUBROUTINE  CLRERR (UNIT)
	INTEGER UNIT

	INTEGER FUNCTION  FCREAD (STRING,ITEMSZ,NITEMS,INUNIT)
	CHARACTER*(*) STRING
	INTEGER ITEMSZ, NITEMS, INUNIT

	INTEGER FUNCTION  FCWRIT (STRING,ITEMSZ,NITEMS,OUTUNIT)
	CHARACTER*(*) STRING
	INTEGER ITEMSZ, NITEMS, OUTUNIT

	INTEGER FUNCTION  FEOF (UNIT)
	INTEGER UNIT

	INTEGER FUNCTION  FERROR (UNIT)
	INTEGER UNIT

	SUBROUTINE  FFLUSH (OUTUNIT)
	INTEGER OUTUNIT

	INTEGER FUNCTION  FGETS (STRING,MAXSTR,INUNIT)
	CHARACTER*(*) STRING
	INTEGER MAXSTR
	INTEGER INUNIT

	INTEGER FUNCTION  FGETC (INUNIT,CH)
	INTEGER INUNIT
	CHARACTER CH

	INTEGER FUNCTION  FPUTC (OUTUNIT,CH)
	INTEGER OUTUNIT
	CHARACTER CH

	SUBROUTINE  FPUTS (STRING,OUTUNIT)
	CHARACTER*(*) STRING
	INTEGER OUTUNIT

	INTEGER FUNCTION  FREAD (BUF,ITEMSZ,NITEMS,INUNIT)
	INTEGER BUF(*)
	INTEGER ITEMSZ, NITEMS, INUNIT

	INTEGER*4 FUNCTION  FSEEK (UNIT,OFFSET,FROM)
	INTEGER UNIT, FROM
	INTEGER*4 OFFSET

	INTEGER*4 FUNCTION  FTELL (UNIT)
	INTEGER UNIT

	INTEGER FUNCTION  FWRITE (BUF,ITEMSZ,NITEMS,OUTUNIT)
	INTEGER BUF(*)
	INTEGER ITEMSZ, NITEMS, OUTUNIT

	INTEGER FUNCTION  GETC (CH)
	CHARACTER CH

	INTEGER FUNCTION  GETS (STRING)
	CHARACTER*(*) STRING

	INTEGER FUNCTION  GETW (INUNIT)
	INTEGER INUNIT

	INTEGER ISATTY (UNIT)
	INTEGER UNIT

	SUBROUTINE  PERROR (STRING)
	CHARACTER*(*) STRING

	INTEGER FUNCTION  PUTC (CH)
	CHARACTER CH

	SUBROUTINE  PUTS (STRING)
	CHARACTER*(*) STRING

	INTEGER FUNCTION  PUTW (WORD,OUTUNIT)
	INTEGER WORD, OUTUNIT

	INTEGER FUNCTION  STRLEN (STRING)
	CHARACTER*(*) STRING

	INTEGER FUNCTION  UNGETC (UNIT,CH)
	INTEGER UNIT
	CHARACTER CH


-------
REMARKS
-------
	     **********************************************
	     *** ALL OF THESE ASSUME THE FILES ARE OPEN ***
	     **********************************************

GETC(), GETS(), PUTC(), and PUTS() correspond to stdin and stdout, which
are normally bound to units 5 and 6 respectively.

Since these get at the internals of 4.1BSD, they are dependent  on  that
release,  and on the availability of source code to obtain the necessary
include file, fio.h.  This is used to obtain the  mapping  from  FORTRAN
unit  number  to C FILE* pointer, encapsulated in the macro CFP() below.
The ordinary FORTRAN formatted I/O library routines  set  various  other
parameters  in  the  control  blocks; init_io() sets a minimal number to
make the routines behave correctly.

These  functions  are implemented here in such a way that normal FORTRAN
I/O READ/WRITE/BACKSPACE statements probably  cannot  be  reliably  used
with them, unless perhaps FFLUSH() is called between them.

By removal of the calls to  init_io(),  and  addition  of  fclose()  and
fopen(),  this  package  could  be  made to use C I/O exclusively.  This
would add considerably to the efficient of fgetc()  and  fputc(),  which
otherwise  must call init_io() each time.  In FORTRAN, files can be read
or written, depending on the first I/O operation performed on them,  and
write  after  read  without an intervening close/open is possible; in C,
access is limited to  the  open  type  (read,  write,  or  append),  and
checking at each I/O system call can be made more limited.


---------------------------------
INTERLANGUAGE CALLING CONVENTIONS
---------------------------------

The 4.xBSD f77 compilers suffix an underscore on external FORTRAN names,
and  both f77 and cc prefix an underscore.  Thus, FORTRAN function foo()
is synonymous with C function  foo_()  and  assembly  language  function
_foo_().

FORTRAN  arguments  are passed by reference, making them equivalent to C
pointers.  Thus,

	INTEGER FUNCTION FOO (A,B,C)
	INTEGER A,B,C

is equivalent to

	int foo_(a,b,c)
	int *a, *b, *c;

FORTRAN SUBROUTINEs are equivalent to void C functions:

	SUBROUTINE FOO (A,B,C)

	void foo_(a,b,c)

For non-CHARACTER functions, the function return mechanism is the  same.
CHARACTER  functions  (which  are not required here) have two additional
initial arguments -- the string address, and its length (which is passed
by VALUE, not reference):

	CHARACTER*17 FUNCTION FOO(a,b,c)
	INTEGER a,b,c

	void foo_(result,len,a,b,c)
	char *result;
	long int len;
	int *a, *b, *c;

Finally,  FORTRAN CHARACTER arguments have their lengths passed by value
as ADDITIONAL arguments at the end of the list:

	SUBROUTINE FOO(A,B,C,N,S)
	INTEGER A,B,N
	CHARACTER*(*) N,S

	void foo_(a,b,c,n,s,lenc,lens)
	int *a,*b,*n;
	char *c,*s;
	long int lenc,lens;

Provided that the FORTRAN I/O library is implemented on top of the C I/O
library, as is commonly the case in Unix implementations, and  that  the
mapping of a FORTRAN unit number to a C FILE* pointer can be determined,
it should be straightforward to adapt these  functions  for  other  Unix
systems.  The only feature likely to cause problems is the mechanism for
passing lengths of  CHARACTER  arguments;  the  scheme  adopted  in  f77
permits  any  data  type to be passed, provided that the called function
does not require the lengths.  HPUX and Convex  FORTRAN  compilers  both
pass   descriptors   for   CHARACTER  variables  instead  of  additional
arguments.

[20-Apr-86] --  Nelson H.F. Beebe, College of Science Computer,
		University of Utah, Salt Lake City, UT 84112
***********************************************************************/

/*!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!*/
/* !!! WARNING !!! CFP() is likely to be implementation-dependent */
#define VALID_UNIT(pn) (((0 <= *pn) && (*pn < MXUNIT)) ? 1 : 0)
#define CFP(pn) (VALID_UNIT(pn) ? ((&units[*pn])->ufd) : ((FILE*) NULL) )
	/* C FILE* pointer for F77 unit *pn */
#define DO_READ 0
#define DO_WRITE 1
/*!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!*/

#include <stdio.h>
#include <strings.h>            /* for definitions of strxxx() */
#include </src/lib/libi77/fio.h> /* for definitions of units[] and ufd */

char* gets();			/* declaration missing in 4.1BSD stdio.h */
int init_io();

/*
	 !!! Maintain ALPHABETICAL function ordering below !!!
*/

void
clrerr_(iounit)
register int* iounit;
{
    (void)clearerr(CFP(iounit));
}


int
fcread_(string,itemsz,nitems,inunit,lenstr)
register char* string;
register int* itemsz;
register int* nitems;
register int* inunit;
register long int lenstr; /* CHARACTER length argument passed by value */
{
    if (init_io(inunit,DO_READ))
	return(fread(string,*itemsz,*nitems,CFP(inunit)));
    else
	return(EOF);
}

int
fcwrit_(string,itemsz,nitems,outunit,lenstr)
register char* string;
register int* itemsz;
register int* nitems;
register int* outunit;
register long int lenstr; /* CHARACTER length argument passed by value */
{
    if (init_io(outunit,DO_WRITE))
	return(fwrite(string,*itemsz,*nitems,CFP(outunit)));
    else
	return(EOF);
}

int
feof_(iounit)
register int* iounit;
{
    return(feof(CFP(iounit)));
}

int
ferror_(iounit)
register int* iounit;
{
    return(ferror(CFP(iounit)));
}

void
fflush_(iounit)
register int *iounit;
{
    (void)fflush(CFP(iounit));
}

int
fgetc_(inunit,ch,lench)
register int *inunit;		/* f77 passes by REFERENCE */
register char *ch;
register long int lench; /* CHARACTER length argument passed by value */
{
    register int c;

    if (!init_io(inunit,DO_READ))
    {
	*ch = '\0';             /* make sure *ch is defined */
	return(EOF);
    }
    c = getc(CFP(inunit));
    if (c < 0)			/* EOF or error */
    {
	*ch = '\0';             /* make sure *ch is defined */
	return (EOF);
    }
    else
    {
	*ch = c;
	return (0);
    }
}

int
fgets_(string,maxstr,inunit,lenstr)
register char *string;		/* f77 passes by REFERENCE */
register int *maxstr;
register int *inunit;
register long int lenstr; /* CHARACTER length argument passed by value */
{
    if (init_io(inunit,DO_READ))
	return ((fgets(string,*maxstr,CFP(inunit)) == (char *)NULL) ? EOF : 0);
    else
	return(EOF);
}

int
fputc_(outunit,ch,lench)
register int *outunit;		/* f77 passes by REFERENCE */
register char *ch;
register long int lench; /* CHARACTER length argument passed by value */
{
    if (init_io(outunit,DO_WRITE))
	return (putc(*ch,CFP(outunit)));
    else
	return(EOF);
}

void
fputs_(string,outunit,lenstr)
register char *string;		/* f77 passes by REFERENCE */
register int *outunit;
register long int lenstr; /* CHARACTER length argument passed by value */
{
    register char c;
    register char* p;

    if (!init_io(outunit,DO_WRITE))
	return;

    p = &string[lenstr];
    c = *p;			/* save terminator character */
    *p = '\0';		/* make valid C string */
    (void)fputs(string,CFP(outunit));
    *p = c;			/* restore terminator character */
}

int
fread_(buf,itemsz,nitems,inunit)
register int* buf;
register int* itemsz;
register int* nitems;
register int* inunit;
{
    if (init_io(inunit,DO_READ))
	return(fread(buf,*itemsz,*nitems,CFP(inunit)));
    else
	return(EOF);
}

long int
fseek_(iounit,offset,from)
register int* iounit;
register long int* offset;
register int* from;
{
    return(fseek(CFP(iounit),*offset,*from));
}

long int
ftell_(iounit)
register int* iounit;
{
    long int k;
    k = ftell(CFP(iounit));
    fprintf(stderr,"ftell(%d) -> %ld\n",*iounit,k);
    return(k);
/*
    return(ftell(CFP(iounit)));
*/
}

int
fwrite_(buf,itemsz,nitems,outunit)
register int* buf;
register int* itemsz;
register int* nitems;
register int* outunit;
{
    if (init_io(outunit,DO_WRITE))
	return(fwrite(buf,*itemsz,*nitems,CFP(outunit)));
    else
	return(EOF);
}

int
getc_(ch,lench)
register char *ch;		/* f77 passes by REFERENCE */
register long int lench;	/* CHARACTER lengths passed by VALUE */
{
    register int c;

    c = getchar();
    if (c < 0)			/* EOF or error */
    {
	*ch = '\0';             /* make sure *ch is defined */
	return (EOF);
    }
    else
    {
	*ch = c;
	return (0);
    }
}

int
gets_(string,lenstr)
register char *string;
register long int lenstr; /* CHARACTER length argument passed by value */
{
    return ((gets(string) == (char *)NULL) ? EOF : 0);
}

int
getw_(inunit)
register int* inunit;
{
    if (init_io(inunit,DO_READ))
	return(getw(CFP(inunit)));
    else
	return(EOF);
}

int
init_io(unit,io_mode)   /* return 1 on success, 0 on failure */
int *unit;
int io_mode;
{
    cilist a;

    a.ciunit = *unit;

    if (!init)
	(void)f_init();		/* initialize FORTRAN I/O system */
    if (c_sfe(&a))		/* do unit checking and initialization */
	return(0);

    curunit = &units[*unit];
    if (io_mode == DO_READ)
    {
	if (curunit->uwrt)      /* if we were writing, change to reading */
	    (void)nowreading(curunit);
    }
    else if (io_mode == DO_WRITE)
    {
	if (!(curunit->uwrt))    /* if we were reading, change to writing */
	    (void)nowwriting(curunit);
    }
    return(1);
}

int
isatty_(iounit)
register int* iounit;
{
    return(isatty(fileno(CFP(iounit))));
}

void
perror_(string,lenstr)
register char *string;
register long int lenstr; /* CHARACTER length argument passed by value */
{
    register char c;
    register char* p;

    p = &string[lenstr];
    c = *p;			/* save terminator character */
    *p = '\0';		/* make valid C string */
    (void)perror(string);
    *p = c;			/* restore terminator character */
}


int
putc_(ch,lench)
register char *ch;		/* f77 passes by REFERENCE */
register long int lench; /* CHARACTER length argument passed by value */
{
    return (putchar(*ch));
}

void
puts_(string,lenstr)
register char *string;
register long int lenstr; /* CHARACTER length argument passed by value */
{
    register char c;
    register char* p;

    p = &string[lenstr];
    c = *p;			/* save terminator character */
    *p = '\0';		/* make valid C string */
    (void)puts(string);
    *p = c;			/* restore terminator character */
}

int
putw_(word,outunit)
register int* word;
register int* outunit;
{
    if (init_io(outunit,DO_WRITE))
	return(putw(*word,CFP(outunit)));
    else
	return(EOF);
}

int
strlen_(string,lenstr)
register char* string;
register long int lenstr;	/* CHARACTER length argument passed by value */
{
    register int k;
    register char c;
    register char* p;

    p = &string[lenstr];
    c = *p;			/* save terminator character */
    *p = '\0';		/* make valid C string */
    k = strlen(string);
    *p = c;			/* restore terminator character */

    return(k);
}

int
ungetc_(inunit,ch,lench)
register int* inunit;
register char* ch;
register long int lench;	/* CHARACTER length argument passed by value */
{
    if (init_io(inunit,DO_READ))
	return(ungetc(*ch,CFP(inunit)));
    else
	return(EOF);
}

