      SUBROUTINE  OUTSTM (IOUNIT,STMT,NS,INDENT,INCONT,WIDTH,BUFFER)
C     (Output Statement)
C     Output a FORTRAN or SFTRAN statement, breaking it correctly
C     over line boundaries if  required.  Comment statements  are
C     also supported.  The arguments are:
C
C     IOUNIT........unit number to contain the output.
C     STMT(*).......statement (packed characters).
C     NS............number of characters in STMT(*).
C     INDENT........current indentation column.
C     INCONT........current indentation column for continued lines.
C     WIDTH.........maximum line width permitted.
C     BUFFER(*).....buffer array sufficiently large to hold one
C                   output line of "WIDTH" characters.
C
C     Only BUFFER(*) is changed  on return.  All other  arguments
C     are preserved.
C
C     The basic problem in outputting multiline indented  FORTRAN
C     or  SFTRAN  statements  is  to  avoid  splitting  character
C     strings.  It is also highly desirable that breaks not occur
C     in variable names.   SUBROUTINE FNDSTR is  used to  isolate
C     character strings (either  quoted or  Hollerith type),  and
C     this information is used to prevent breaking a string  over
C     a line boundary.  The algorithm  is based on the  following
C     rules:
C
C     1..If a  break is required  outside  of a character string,
C        break after one of the characters " =+-/)*", except that
C        "**" must not be  broken in the middle,  and "+" or  "-"
C        may be not preceded by an exponent letter ("E", "D",  or
C        "Q").   A   break   cannot   be   permitted   inside   a
C        multi-character comment sentinel.  If no break is found,
C        then try to  break after a  comma (which might  separate
C        array subscripts).   If this  fails, then  try to  break
C        before a  left parenthesis.   If  this too  fails,  then
C        force an arbitrary  break after the  last column of  the
C        output line.
C
C     2..If a break would be required inside a character string,
C        then start a new line and check rule 3.
C
C     3..If the character string fits on the line after the
C        current indentation column, put it there.  Otherwise
C        check rule 4.
C
C     4..If the character string can be fit on the line by moving
C        the indentation column left, do so by placing the string
C        as far to the  right as possible, so  as to perturb  the
C        indentation as little as possible.  Otherwise check rule
C        5.
C
C     5..If the character string cannot fit on the  current line,
C        then temporarily reset the  indentation to column 7  and
C        output the string  on as many  lines as required.   Then
C        restore the indentation to its original value.
C
C     6..Whenever a character string is  placed on a line,  leave
C        at least one space following it (except if rule 5 is  in
C        effect), since it will usually be followed by one of the
C        terminators " /,)".
C
C     None of these  rules need  be considered  if the  statement
C     fits on a  single output line,  and in such  a case, it  is
C     even examined to see if it contains any character  strings.
C     Similarly, statements which can fit  on one line are  never
C     squished to  remove non-significant  blanks.  This  ensures
C     that user-intended  spacing in  one-liners remains  intact.
C     If the text requires more  than output line, then  multiple
C     blanks are squished to single  blanks.  This of course  can
C     mess  up  FORMAT  and  DATA  statements,  but  does  permit
C     prettyprinting source text which has already been indented,
C     and therefore  contains long  embedded strings  of  blanks.
C     Because character strings are split only if they exceed the
C     line length, the algorithm  actually does a reasonable  job
C     on FORMAT statements.
C
C     (30-AUG-83)
C  - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
C
C     EXTERNAL REFERENCES (FUNCTION,SUBROUTINE,COMMON)
C
C     EXTERNAL REFS       CHGET3,      F4CLC,       F4MVC,       FNDSTR
C     EXTERNAL REFS       MIN0,        WRITLN
C
C  - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
C
C     EXTERNAL FUNCTIONS AND SUBROUTINES
C
      INTEGER             CHGET3
      LOGICAL             F4CLC
C  - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
C
C     INTRINSIC FUNCTIONS
C
      INTEGER             MIN0
C  - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
C
C     STATEMENT FUNCTIONS
C
      INTEGER             NWDSF,       REST
      LOGICAL             DECDIG,      LETTER
C  - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
C
C     NON-COMMON VARIABLES
C
C     SETS                ACHAR,       BFCONT,      BFLAST,      BFSAVE
C     SETS                BFSTRT,      CCBRK,       CCL,         CCLOOK
C     SETS                CCMIN,       CCNOW,       CCR,         COMARZ
C     SETS                COMENT,      DRZ,         EQULRZ,      ERZ
C     SETS                INLOOP,      LASTCH,      LENSTR,      MAXIND
C     SETS                NCOPY,       NICE,        NSTMT,       QRZ
C     SETS                SQUISH,      STRING
C
      INTEGER             ACHAR,       BFCONT,      BFLAST,      BFSAVE
      INTEGER             BFSTRT,      BUFFER(1),   CCBRK,       CCL
      INTEGER             CCLOOK,      CCMIN,       CCNOW,       CCR
      INTEGER             COMARZ,      DRZ,         EQULRZ,      ERZ
      INTEGER             IARG,        ILAST,       INCONT,      INDENT
      INTEGER             INOW,        IOUNIT,      LASTCH,      LENSTR
      INTEGER             MAXIND,      NCOPY,       NS,          NSTMT
      INTEGER             QRZ,         STMT(1),     WIDTH
      LOGICAL             COMENT,      ERROR,       INLOOP,      NICE
      LOGICAL             QUOTED,      SQUISH,      STRING
C
C  - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
C
C     COMMON BLOCK / COM04  /
C
      INTEGER             ARZ,         BLNKRZ,      COLNRZ,      CRZ
      INTEGER             DASHRZ,      HRZ,         LEFTRZ,      NINERZ
      INTEGER             PLUSRZ,      QUOTRZ,      RITERZ,      SLSHRZ
      INTEGER             STARRZ,      ZERORZ,      ZRZ
C
      COMMON / COM04  /   ARZ,         BLNKRZ,      COLNRZ,      CRZ
      COMMON / COM04  /   DASHRZ,      HRZ,         LEFTRZ,      NINERZ
      COMMON / COM04  /   PLUSRZ,      QUOTRZ,      RITERZ,      SLSHRZ
      COMMON / COM04  /   STARRZ,      ZERORZ,      ZRZ
C
C  - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
C
C     COMMON BLOCK / COM11  /
C
      INTEGER             COMCHR,      LCCHR
C
      COMMON / COM11  /   COMCHR(1),   LCCHR
C
C  - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
C
C     COMMON BLOCK / COM14  /
C
      INTEGER             WDSIZE
C
      COMMON / COM14  /   WDSIZE
C
C
      DATA                MAXIND    / 57 /
C
C
C     LOGICAL statement function returning .TRUE. if right-zero
C     character argument is a decimal digit.
C
      DECDIG(IARG) = (ZERORZ .LE. IARG) .AND. (IARG .LE. NINERZ)
C
C     LOGICAL statement function returning .TRUE. if right-zero
C     character argument is a letter.
C
      LETTER(IARG) = (ARZ .LE. IARG) .AND. (IARG .LE. ZRZ)
C
C     Statement function defining number of words in a string  of
C     IARG characters.  This function  is used in ADDHT,  IOFUNC,
C     and OUTSTM, and requires WDSIZE defined in /COM14/.
C
      NWDSF(IARG) = (IARG+WDSIZE-1)/WDSIZE
C
C
C     REST is  the number  of characters  in a  string after  the
C     current character, INOW, which has already been  processed.
C
      REST(INOW,ILAST) = ILAST  - INOW
      ASSIGN 20001 TO NPR001
      GO TO 30001
20001 CONTINUE
20002 IF (.NOT.(CCNOW .LT. NSTMT)) GO TO 20003
      ASSIGN 20004 TO NPR002
      GO TO 30002
20004 ASSIGN 20005 TO NPR003
      GO TO 30003
20005 IF (.NOT.(LENSTR .GT. 0)) GO TO 20006
      ASSIGN 20009 TO NPR004
      GO TO 30004
20009 CONTINUE
20006 CONTINUE
20007 GO TO 20002
20003 ASSIGN 20010 TO NPR005
      GO TO 30005
C
20010 RETURN
C
C-----------------------------------------------------------------------
C---- PROCEDURE (Advance Over Leading Blanks)
30006 CONTINUE
20011 IF (.NOT.((CCNOW .LT. NSTMT) .AND. (CHGET3(STMT,CCNOW+1)
     X.EQ. BLNKRZ))) GO TO 20012
      CCNOW = CCNOW + 1
      GO TO 20011
20012 GO TO NPR006, (20043,20101)
C
C-----------------------------------------------------------------------
C---- PROCEDURE (Clear Output Buffer)
30007 CALL F4MVC (1H ,1,BUFFER,1,1)
      CALL F4MVC (BUFFER,1,BUFFER,2,WIDTH-1)
      BFLAST = BFSTRT
      GO TO NPR007, (20094,20107)
C
C-----------------------------------------------------------------------
C---- PROCEDURE (Copy Character String Into Output Buffer)
30004 IF (.NOT.(LENSTR .LT. REST(BFLAST,WIDTH))) GO TO 20013
      ASSIGN 20016 TO NPR008
      GO TO 30008
20016 GO TO 20014
20013 ASSIGN 20017 TO NPR005
      GO TO 30005
20017 IF (.NOT.(LENSTR .LT. REST(BFLAST,WIDTH))) GO TO 20018
      ASSIGN 20021 TO NPR008
      GO TO 30008
20021 GO TO 20019
20018 IF (.NOT.(LENSTR .LT. REST(6,WIDTH))) GO TO 10001
      BFLAST = WIDTH - LENSTR - 1
      ASSIGN 20022 TO NPR008
      GO TO 30008
20022 GO TO 20019
10001 BFSAVE = BFCONT
      BFCONT = 6
      ASSIGN 20023 TO NPR008
      GO TO 30008
20023 BFCONT = BFSAVE
20019 CONTINUE
20020 CONTINUE
20014 GO TO 20009
C
C-----------------------------------------------------------------------
C---- PROCEDURE (Copy Characters Preceding String Into Output Buffer)
30003 NCOPY = REST(CCNOW,CCL-1)
      INLOOP = .FALSE.
20024 IF (.NOT.(NCOPY .GT. 0)) GO TO 20025
      IF (.NOT.(NCOPY .GT. REST(BFLAST,WIDTH))) GO TO 20026
      ASSIGN 20029 TO NPR009
      GO TO 30009
20029 IF (.NOT.(NICE .OR. INLOOP)) GO TO 20030
      NCOPY = REST(CCNOW,CCBRK)
      GO TO 20031
20030 ASSIGN 20033 TO NPR005
      GO TO 30005
20033 INLOOP = .TRUE.
      GO TO 20024
20031 CONTINUE
20032 CONTINUE
20026 IF (.NOT.(SQUISH)) GO TO 20034
      ASSIGN 20037 TO NPR010
      GO TO 30010
20037 GO TO 20035
20034 ASSIGN 20038 TO NPR011
      GO TO 30011
20038 CONTINUE
20035 IF (.NOT.(BFLAST .GE. WIDTH)) GO TO 20039
      ASSIGN 20042 TO NPR005
      GO TO 30005
20042 CONTINUE
20039 IF (.NOT.(SQUISH)) GO TO 20043
      ASSIGN 20043 TO NPR006
      GO TO 30006
20043 NCOPY = REST(CCNOW,CCL-1)
      INLOOP = .FALSE.
      GO TO 20024
20025 GO TO 20005
C
C-----------------------------------------------------------------------
C---- PROCEDURE (Copy String Into Buffer)
30008 CONTINUE
20044 IF (.NOT.(LENSTR .GT. 0)) GO TO 20045
      NCOPY = MIN0(LENSTR,REST(BFLAST,WIDTH))
      IF (.NOT.(NCOPY .LE. 0)) GO TO 20046
      ASSIGN 20049 TO NPR005
      GO TO 30005
20049 NCOPY = MIN0(LENSTR,REST(BFLAST,WIDTH))
20046 CALL F4MVC (STMT,CCL,BUFFER,BFLAST+1,NCOPY)
      BFLAST = BFLAST + NCOPY
      CCL = CCL + NCOPY
      LENSTR = REST(CCL-1,CCR)
      GO TO 20044
20045 CCNOW = CCR
      GO TO NPR008, (20016,20021,20022,20023)
C
C-----------------------------------------------------------------------
C---- PROCEDURE (Copy Text Removing Non-Significant Blanks)
30010 ACHAR = CHGET3(STMT,CCNOW+1)
      CCNOW = CCNOW + 1
      NCOPY = NCOPY - 1
      CALL F4MVC (STMT,CCNOW,BUFFER,BFLAST+1,1)
      BFLAST = BFLAST + 1
20050 IF (.NOT.(NCOPY .GT. 0)) GO TO 20051
      LASTCH = ACHAR
      ACHAR = CHGET3(STMT,CCNOW+1)
      CCNOW = CCNOW + 1
      NCOPY = NCOPY - 1
      IF (.NOT.(ACHAR .NE. BLNKRZ)) GO TO 20052
      CALL F4MVC (STMT,CCNOW,BUFFER,BFLAST+1,1)
      BFLAST = BFLAST + 1
      GO TO 20053
20052 IF (.NOT.(LASTCH .NE. BLNKRZ)) GO TO 10002
      CALL F4MVC (STMT,CCNOW,BUFFER,BFLAST+1,1)
      BFLAST = BFLAST + 1
      GO TO 20053
10002 CONTINUE
20053 GO TO 20050
20051 GO TO 20037
C
C-----------------------------------------------------------------------
C---- PROCEDURE (Copy Text Verbatim)
30011 CALL F4MVC (STMT,CCNOW+1,BUFFER,BFLAST+1,NCOPY)
      BFLAST = BFLAST + NCOPY
      CCNOW = CCNOW + NCOPY
      GO TO 20038
C
C-----------------------------------------------------------------------
C---- PROCEDURE (Find Next String)
30002 IF (.NOT.(STRING)) GO TO 20055
      CALL FNDSTR (STMT,CCNOW+1,NSTMT,CCL,CCR,QUOTED,ERROR)
      IF (.NOT.(CCR .GT. CCL)) GO TO 20058
      LENSTR = REST(CCL-1,CCR)
      GO TO 20059
20058 STRING = .FALSE.
      LENSTR = 0
      CCL = NSTMT + 1
20059 CONTINUE
20055 GO TO 20004
C
C-----------------------------------------------------------------------
C---- PROCEDURE (Find Place To Break String)
30009 NICE = .TRUE.
      CCLOOK = CCNOW + REST(BFLAST,WIDTH) + 1
      CCBRK = 0
      CCMIN = CCNOW + 1
20061 IF (.NOT.((CCBRK .EQ. 0) .AND. (CCLOOK .GT. CCMIN))) GO TO 20062
      CCLOOK = CCLOOK - 1
      ACHAR = CHGET3(STMT,CCLOOK)
      IF (.NOT.(LETTER(ACHAR))) GO TO 20063
      GO TO 20064
20063 IF (.NOT.(DECDIG(ACHAR))) GO TO 10003
      GO TO 20064
10003 IF (.NOT.(ACHAR .EQ. BLNKRZ)) GO TO 10004
      CCBRK = CCLOOK
      GO TO 20064
10004 IF (.NOT.(ACHAR .EQ. EQULRZ)) GO TO 10005
      CCBRK = CCLOOK
      GO TO 20064
10005 IF (.NOT.(ACHAR .EQ. SLSHRZ)) GO TO 10006
      CCBRK = CCLOOK
      GO TO 20064
10006 IF (.NOT.(ACHAR .EQ. RITERZ)) GO TO 10007
      CCBRK = CCLOOK
      GO TO 20064
10007 IF (.NOT.(ACHAR .EQ. PLUSRZ)) GO TO 10008
      ACHAR = CHGET3(STMT,CCLOOK-1)
      IF ((ACHAR .NE. ERZ) .AND. (ACHAR .NE. DRZ) .AND.
     X(ACHAR .NE. QRZ)) CCBRK = CCLOOK
      GO TO 20064
10008 IF (.NOT.(ACHAR .EQ. DASHRZ)) GO TO 10009
      ACHAR = CHGET3(STMT,CCLOOK-1)
      IF ((ACHAR .NE. ERZ) .AND. (ACHAR .NE. DRZ) .AND.
     X(ACHAR .NE. QRZ)) CCBRK = CCLOOK
      CCBRK = CCLOOK
      GO TO 20064
10009 IF (.NOT.(ACHAR .EQ. STARRZ)) GO TO 10010
      IF (.NOT.(CHGET3(STMT,CCLOOK+1) .NE. STARRZ)) GO TO 20066
      CCBRK = CCLOOK
20066 CONTINUE
10010 CONTINUE
20064 IF (.NOT.(CCBRK .GT. 0)) GO TO 20069
      IF (.NOT.(LCCHR .GT. 1)) GO TO 20072
      IF (F4CLC(COMCHR,1,STMT,CCBRK,LCCHR)) CCBRK =
     X0
20072 CONTINUE
20069 GO TO 20061
20062 IF (.NOT.(CCBRK .EQ. 0)) GO TO 20075
      CCLOOK = CCNOW + REST(BFLAST,WIDTH) + 1
20078 IF (.NOT.((CCBRK .EQ. 0) .AND. (CCLOOK .GT. CCMIN))) GO TO 20079
      CCLOOK = CCLOOK - 1
      ACHAR = CHGET3(STMT,CCLOOK)
      IF (ACHAR .EQ. COMARZ) CCBRK = CCLOOK
      GO TO 20078
20079 CONTINUE
20075 IF (.NOT.(CCBRK .EQ. 0)) GO TO 20080
      CCLOOK = CCNOW + REST(BFLAST,WIDTH) + 1
      CCMIN = CCMIN + 1
20083 IF (.NOT.((CCBRK .EQ. 0) .AND. (CCLOOK .GT. CCMIN))) GO TO 20084
      CCLOOK = CCLOOK - 1
      ACHAR = CHGET3(STMT,CCLOOK)
      IF (ACHAR .EQ. LEFTRZ) CCBRK = CCLOOK - 1
      GO TO 20083
20084 CONTINUE
20080 IF (.NOT.(CCBRK .EQ. 0)) GO TO 20085
      CCBRK = CCNOW + REST(BFLAST,WIDTH)
      IF (.NOT.(LCCHR .GT. 1)) GO TO 20088
      IF (F4CLC(COMCHR,1,STMT,CCBRK,LCCHR)) CCBRK =
     XCCBRK - 1
20088 NICE = .FALSE.
20085 GO TO 20029
C
C-----------------------------------------------------------------------
C---- PROCEDURE (Initialize Output Buffer)
30012 IF (.NOT.(COMENT)) GO TO 20091
      BFSTRT = 0
      GO TO 20092
20091 BFSTRT = MIN0(MAXIND,INDENT) - 1
20092 BFCONT = MIN0(MAXIND,INCONT) - 1
      ASSIGN 20094 TO NPR007
      GO TO 30007
20094 CALL F4MVC (STMT,1,BUFFER,1,MIN0(NSTMT,6))
      IF (COMENT) BFLAST = MIN0(NSTMT,6)
      GO TO 20097
C
C-----------------------------------------------------------------------
C---- PROCEDURE (Initialize RJZF Constants)
30013 COMARZ = CHGET3(1H,,1)
      DRZ = CHGET3(1HD,1)
      ERZ = CHGET3(1HE,1)
      QRZ = CHGET3(1HQ,1)
      EQULRZ = CHGET3(1H=,1)
      GO TO 20096
C
C-----------------------------------------------------------------------
C---- PROCEDURE (Initialize)
30001 ASSIGN 20095 TO NPR014
      GO TO 30014
20095 ASSIGN 20096 TO NPR013
      GO TO 30013
20096 COMENT = F4CLC(1HC,1,STMT,1,1) .OR. F4CLC(1H*,1,STMT,1,1)
      ASSIGN 20097 TO NPR012
      GO TO 30012
20097 CCNOW = 6
      IF (.NOT.(COMENT)) GO TO 20098
      STRING = .FALSE.
      SQUISH = .FALSE.
      GO TO 20099
20098 ASSIGN 20101 TO NPR006
      GO TO 30006
20101 STRING = REST(INDENT-1,WIDTH) .LT. REST(CCNOW,NSTMT)
      SQUISH = STRING
20099 CCL = NSTMT + 1
      CCR = 0
      LENSTR = 0
      GO TO 20001
C
C-----------------------------------------------------------------------
C---- PROCEDURE (Trim Trailing Blanks)
30014 NSTMT = NS
20102 IF (.NOT.((CHGET3(STMT,NSTMT) .EQ. BLNKRZ) .AND. (NSTMT .GT.
     X1))) GO TO 20103
      NSTMT = NSTMT - 1
      GO TO 20102
20103 GO TO 20095
C
C-----------------------------------------------------------------------
C---- PROCEDURE (Write Current Output Buffer)
30005 IF (.NOT.(BFLAST .GT. BFSTRT)) GO TO 20104
      CALL WRITLN (IOUNIT,BUFFER,NWDSF(BFLAST))
      BFSTRT = BFCONT
      ASSIGN 20107 TO NPR007
      GO TO 30007
20107 IF (.NOT.(COMENT)) GO TO 20108
      CALL F4MVC (1HC,1,BUFFER,1,1)
      GO TO 20109
20108 CALL F4MVC (1HX,1,BUFFER,6,1)
20109 CONTINUE
20104 GO TO NPR005, (20010,20017,20033,20042,20049)
C
      END
